/* 
 *    SS3DSoak - SSE, SSE2, 3DNow Reliability Test
 *    Copyright (C) Roy Longbottom 2002 - 2004 
 *
 *  This uses the much the same background code as SSE3DNow -
 *  See description in SSE3DNow.cpp
 *
 *  See also SS3DSoak.txt for detail of specific tests.
 *
 *  Dialog box TestPanel is used to select which test is run.   
 *  The main window is regularly updated with a single progress
 *  line and SoakLog.txt every minute.
 *
 *  Three sets of tests are provided CPU Only with minimum data
 *  access, L1 cache/CPU and Cache/RAM. Up to 3 run buttons are
 *  produced for each in TestPanel subject to SSE, SSE2 or 3DNow
 *  instructions being available. The following button IDs are
 *  in TestPanel():
 * 
 *  SSE_CPU, SSE2_CPU, DNOW_CPU
 *  SSE_CACHE, SSE2_CACHE, DNOW_CACHE 
 *  SSE_RAM, SSE2_RAM, DNOW_RAM
 *
 *  These set a testFunction and phase = 1 to execute setUp() via
 *  WinMain then runTests() at phase = 2.
 *
 *  AT_KB selects memory size used for L1 cache tests
 *  AT_MB selects memory size used for cache/RAM tests
 *  AT_MINS selects running time minutes
 *  EXIT_BUTTON
 *
 *  runTests() executes one of the following which include the
 *  appropriate assembly code (testFunction 0 to 8):
 *
 *  SSEAM(), SSE2AM(), CPU3DNowAM
 *  CacheSSE1(), CacheSSE2(), Cache3DNow()
 *  RAMSSE1(), RAMSSE2(), RAM3DNow()
 *
 *  The first six are also run via setUp() to obtain numeric results
 *  for comparison and for run time calibration.
 */
 
#define NAME "Test"
#define TITLE " SSE, SSE2 and 3DNow Reliability Test"
#define INITGUID

#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <xmmintrin.h>
#include <malloc.h>
#include <math.h>
#include "SS3DSoak.h"

//  ************** PROTOPTYPES *****************

BOOL FAR PASCAL AboutDlgProc( HWND, UINT, WPARAM, LPARAM);
static BOOL InitApp(HINSTANCE, LPSTR, int);
long FAR PASCAL WindowProc(HWND, UINT, WPARAM, LPARAM);                  
static void CleanUpAndPostQuit(void);
void displayWhatever(HWND);

BOOL FAR PASCAL TestPanel( HWND, UINT, WPARAM, LPARAM);

void local_time();
void end_time();
void start_time();

BOOL setUp();
BOOL finish();
BOOL runTests();
BOOL CPU3DNowAM();
BOOL Cache3DNow();
BOOL RAM3DNow();
BOOL SSEAM();
BOOL CacheSSE1();
BOOL RAMSSE1();
BOOL SSE2AM();
BOOL CacheSSE2();
BOOL RAMSSE2();
BOOL saveNumbers();

// ************* GLOBAL VARIABLES **********

float  *spRand;
double *dpRand; 
float  *spData;
double *dpData;
float  *spResults;
double *dpResults;
float  spCompare[1024];
double dpCompare[1024];

LARGE_INTEGER starttime;

HWND     mainHWND;
HWND     testHwnd = NULL;
HINSTANCE   MyInstance;
RECT     winRect;
FILE     *outfile;

BOOL   bQuit = FALSE;
BOOL   errorFound = FALSE;
BOOL   autoRun = FALSE;
BOOL   doTesting = FALSE;
BOOL   flash;
BOOL   runDone = FALSE;
BOOL   end;
BOOL   isSSE2 = FALSE;
BOOL   started = FALSE;
BOOL   firstPass;
BOOL   saveRand = FALSE;
BOOL   saveData = FALSE;
BOOL   multiWin = FALSE;

short  keyPress;

int  passesDone;
int  ramKB;
int  cacheRun = 0; // 4 KB;
int  RAMRun = 10;  // 4 MB;
int  KBsizes = 11;
int  MBsizes = 19;
int  KBCache[50];
int  KBMemory[50];
int  useCacheKB = 4;
int  useRAMKB   = 4096;
int  timeToDo = 1;
int  runMins = 5;
int  MaxMins = 10;
int  maxTests = 3; 
int  screenWide;
int  screenHigh;
int  testFunction = 0;
int  phase = 0;
int  errorCount;
int  theseErrors;
int  runMult;
int  ramTest;
int  ramMult;
int  runTimes[20];
int  xpos = 0;
int  ypos = 0;

char  version[30] = "Version 1.0";
char  logfile[80] = "SoakLog.txt";
char  timeday[30];
char  writeMsg[256];
char  writeMsg2[256];
char  runMsg[256] = " ";
char  errorMsg[256] = " ";
char  testMsg[256] = " ";
char  windowText[256];
char  commandLine[1024] = "";
char  runNames[10][20];
char  startRam[200];
char  timeMsg[100];
char  speedMsg[100];

double  runningSecs;
double  loopSecs; 
double  theseSecs;
double  theseMB;
double  countMB;
double  theseMFops;
double  countMFops;
double  memKWords = 1;


void displayWhatever(HWND hwnd)
{
    sprintf(windowText, " SSE, SSE2 and 3DNow Reliability Tests");
    SetWindowText(mainHWND, windowText);        
        
    if (doTesting)
    {
        if (testHwnd)   DestroyWindow(testHwnd);
    }
    else
    {    
        testHwnd = CreateDialog( MyInstance,"DIALOG_8", hwnd, (DLGPROC)TestPanel);
    }  
} // displayWhatever


// WinMain - initialization, message loop

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR cmdline, int cmdshow)
{
    MSG         msg;
    
    hPrevInstance = hPrevInstance;

    if (!InitApp(hInstance, cmdline, cmdshow))
         return FALSE;

    while(!bQuit)
    {
        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
        {
            
           if (msg.message == WM_QUIT)
           {
                CleanUpAndPostQuit();
                break;
           }
   
           if( !GetMessage( &msg, NULL, 0, 0 ) ) return msg.wParam;
           TranslateMessage(&msg);
           DispatchMessage(&msg);
        }
        else if (phase == 1)
        {
            if (!setUp())
            {
                CleanUpAndPostQuit();
                phase = 0;
            }
            phase = 2;
        }
        else if (phase == 2)
        {
            if (!runTests())
            {
                fflush(outfile);
                fclose(outfile);
                CleanUpAndPostQuit();
                phase = 0;
            }
        }
        else if (phase == 3)
        {
            runDone = TRUE;
            if (!finish())
            {
                CleanUpAndPostQuit();
                phase = 0;
            }
            phase = 0;
            runDone = TRUE;
            if (autoRun) CleanUpAndPostQuit();
        }
        else
        {
           // make sure we go to sleep if we have nothing else to do
           WaitMessage();
        }
    }
    return msg.wParam;
    
} // WinMain

// ***************************************************************************
//                    Initialization and object creation                     *
// ***************************************************************************

// InitApp
// Creates window

static BOOL
InitApp(HINSTANCE hInstance, LPSTR cmdline, int cmdshow)
{
    HWND hwnd;
    WNDCLASS wc;
    int   mult = 20;
    int   i, j;
    char *delims = {" ,"};
    char commands[20][256];
    char *p;
    int  KBytes = 0;
    char buf[256] = "";
    HDC  hdc;
    int minutes = 0;
    int newPos =0;
    int valid = 0;
    double  dval;
  
    
    // set up and registers the window class
    
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, "GenericIcon");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName = "GenericMenu";
    wc.lpszClassName = NAME;
    if (!RegisterClass(&wc))
        return FALSE;

    // Create the window

    hwnd =
        CreateWindow(
            NAME,
            TITLE,
            WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL,
            0,
            0,
            640,
            480,
            NULL, 
            NULL, 
            hInstance, 
            NULL 
        );

    if (!hwnd) return FALSE;
    mainHWND = hwnd;
    strcpy(commandLine, strupr(cmdline));

    hdc = GetDC(mainHWND);
    screenWide = GetDeviceCaps(hdc, HORZRES);
    screenHigh = GetDeviceCaps(hdc, VERTRES);   
    ReleaseDC(mainHWND, hdc);
    
    GetClientRect(mainHWND, &winRect);

    WhatConfig();
    strcpy(startRam, configData3);
    ramKB = ramsize / 1024;

    if (!hasSSE && !hasSSE2 && !has3DNow)
    {
        MessageBox(NULL, "No SSE, SSE2 or 3DNow instructions available", "ERROR", MB_ICONWARNING | MB_OK);
        CleanUpAndPostQuit();
        return FALSE;
    }
    runTimes[0] = 1;
    runTimes[1] = 5;
    runTimes[2] = 10;
    runTimes[3] = 30;
    runTimes[4] = 60;
    runTimes[5] = 120;
    runTimes[6] = 240;
    runTimes[7] = 480;
    runTimes[8] = 960;
    runTimes[9] = 1440;

    KBCache[0]  = 4;
    KBCache[1]  = 8;
    KBCache[2]  = 16;
    KBCache[3]  = 32;
    KBCache[4]  = 64;
    KBCache[5]  = 128;
    KBCache[6]  = 256;
    KBCache[7]  = 512;
    KBCache[8]  = 1024;
    KBCache[9]  = 2048;
    KBCache[10] = 4096;

    KBMemory[0]  = 4;
    KBMemory[1]  = 8;
    KBMemory[2]  = 16;
    KBMemory[3]  = 32;
    KBMemory[4]  = 64;
    KBMemory[5]  = 128;
    KBMemory[6]  = 256;
    KBMemory[7]  = 512;
    KBMemory[8]  = 1024;
    KBMemory[9]  = 2048;
    KBMemory[10] = 4096;
    KBMemory[11] = 8192;
    KBMemory[12] = 16384;
    KBMemory[13] = 32768;
    KBMemory[14] = 65536;
    KBMemory[15] = 131072;
    KBMemory[16] = 262144;
    KBMemory[17] = 524288;
    KBMemory[18] = 1048576;
    KBMemory[19] = 2097152;
    KBMemory[20] = 4194304;

    i = 0;
    p = strtok(commandLine, delims);
    while (p != NULL)
    {
        strcpy(commands[i], p);
        strcat(buf, commands[i]);
        strcat(buf, " ");
        i = i + 1;
        if (i > 9) break;
        p = strtok(NULL, delims);
    }
    if (i > 0)
    {
        strcpy(commandLine, buf);
    }
    else
    {
        strcpy(commandLine, "None");
    }
    autoRun = FALSE;
       
    for (j=0; j<i; j++)
    {
        switch(commands[j][0])
        {            
            case 'A': // Auto Run
             autoRun = TRUE;
             phase = 1;
             break;            

            case 'K': // KBytes
             if (j < i -1)
             {
                 sscanf(commands[j+1], "%d", &KBytes);
                 if (KBytes < 4) KBytes = 4;
             }
             break;            
       
            case 'M': // Minutes
             if (j < i -1)
             {
                 sscanf(commands[j+1], "%d", &minutes);
                 if (minutes < 1) minutes = 1;
             }
             break;

            case 'L': // log file
             if (j < i -1)
             {
                 sscanf(commands[j+1], "%s", logfile);
             }
             break;
        }            
    }
    for (j=0; j<i; j++)
    {
        if (strnicmp(commands[j],  "SSE", 3) == 0)
        {
            if (hasSSE && hasSSEOS)
            {
                valid = valid + 1;
                isSSE2 = FALSE;
            }
            else
            {
                MessageBox(NULL, "SSE N/A", "COMMAND ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
                CleanUpAndPostQuit();
                return FALSE;
            }
        }
        if (strnicmp(commands[j],  "SSE2", 4) == 0)
        {
            if (hasSSE2 && hasSSE2OS)
            {
                testFunction = testFunction + 1;
                valid = valid + 1;
                isSSE2 = TRUE;
            }
            else
            {
                MessageBox(NULL, "SSE2 N/A", "COMMAND ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
                CleanUpAndPostQuit();
                return FALSE;
            }
        }
        if (strnicmp(commands[j],  "3DNOW", 5) == 0)
        {
            if (has3DNow)
            {
                testFunction = testFunction + 2;
                valid = valid + 1;
                isSSE2 = FALSE;
            }
            else
            {
                MessageBox(NULL, "3DNow N/A", "COMMAND ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
                CleanUpAndPostQuit();
                return FALSE;
            }
        }
        if (strnicmp(commands[j],  "CPU", 3) == 0)
        {
            valid = valid + 10;
        }
        if (strnicmp(commands[j],  "CACHE", 5) == 0)
        {
            testFunction = testFunction + 3;
            valid = valid + 10;
        }
        if (strnicmp(commands[j],  "RAM", 3) == 0)
        {
            testFunction = testFunction + 6;
            valid = valid + 10;
        }
        if (strnicmp(commands[j],  "P1", 2) == 0)
        {
            xpos = 0;
            ypos = 0;
            multiWin = TRUE;
        }
        if (strnicmp(commands[j],  "P2", 2) == 0)
        {
            xpos = 0;
            ypos = screenHigh - 500;
            if (ypos < 0) ypos = 0;
            multiWin = TRUE;
        }
        if (strnicmp(commands[j],  "P3", 2) == 0)
        {
            xpos = screenWide - 660;
            if (xpos < 0) xpos = 0;
            ypos = 0;
            multiWin = TRUE;
        }
        if (strnicmp(commands[j],  "P4", 2) == 0)
        {
            xpos = screenWide - 660;
            if (xpos < 0) xpos = 0;
            ypos = screenHigh - 500;
            if (ypos < 0) ypos = 0;
            multiWin = TRUE;
        }
        if (strnicmp(commands[j],  "SAVERAND", 8) == 0)
        {
            saveRand = TRUE;
        }
        if (strnicmp(commands[j],  "SAVEDATA", 8) == 0)
        {
            saveData = TRUE;
        }
    }
    if (i > 0 && valid != 11)
    {
         sprintf(writeMsg, " %d Invalid", valid);
         MessageBox(NULL, writeMsg, "COMMAND ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
         CleanUpAndPostQuit();
         return FALSE;
    }         
    if (minutes > 0)
    {
        for (i=0; i<MaxMins; i++)
        {
            timeToDo = i;
            runMins  = runTimes[i];
            if (minutes == runMins) i = MaxMins;
            if (minutes <  runMins) i = MaxMins;
        }
    }
    if (testFunction < 3)
    {
        memKWords = 1;
    }
    else if (testFunction < 6)
    {
        if (KBytes > 0)
        {
            for (i=0; i<KBsizes; i++)
            {
                cacheRun = i;
                useCacheKB = KBCache[i];
                if (KBytes == useCacheKB) i = KBsizes;
                if (KBytes <  useCacheKB) i = KBsizes;             
            }
        }
        memKWords = useCacheKB / 4;
        if (isSSE2) memKWords = memKWords / 2;
    }
    else
    {
        if (KBytes > 0)
        {
            for (i=0; i<MBsizes; i++)
            {
                RAMRun = i;
                useRAMKB = KBMemory[i];
                if (KBytes == useRAMKB) i = MBsizes;
                if (KBytes <  useRAMKB) i = MBsizes;
            }
            if (useRAMKB > ramKB - 32768)
            {
                sprintf(writeMsg, " %d KB memory selected with memory size %d KB", useRAMKB, ramKB);
                MessageBox(NULL, writeMsg, "WARNING OF PAGING TO DISK", MB_ICONWARNING | MB_OK);
            }
        }
        memKWords = useRAMKB / 4;
        if (isSSE2) memKWords = memKWords / 2;
    }       

    spRand = (float *)_aligned_malloc(4096, 16);
    if (spRand == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        CleanUpAndPostQuit();
        return FALSE;
    }

    dpRand = (double *)_aligned_malloc(8192, 16);
    if (dpRand == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        _aligned_free(spRand);
        CleanUpAndPostQuit();
        return FALSE;
    }
    dpResults = (double *)_aligned_malloc(8192, 16);
    if (dpResults == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        _aligned_free(spRand);
        _aligned_free(dpRand);
        CleanUpAndPostQuit();
        return FALSE;
    }
    spResults = (float *)_aligned_malloc(4096, 16);
    if (spResults == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        _aligned_free(spRand);
        _aligned_free(dpRand);
        _aligned_free(dpResults);
        CleanUpAndPostQuit();
        return FALSE;
    }

    started = TRUE;

    srand(456);
    for (i=0; i<1024; i++)
    {
        dval = pow(2, ((240 * ((double)rand() / 32768)) - 120))
                                  *  (double)rand() / 16384;
        spRand[i] =  (float)dval; 
        dpRand[i] =  pow(2, ((2020 * ((double)rand() / 32768)) - 1010))
                                   *  (double)rand() / 16384;
    }
    if (saveRand) saveNumbers();
    
    sprintf(runNames[0], "SSE CPU Test");
    sprintf(runNames[1], "SSE2 CPU Test");
    sprintf(runNames[2], "3DNow CPU Test");
    sprintf(runNames[3], "SSE Cache Test");
    sprintf(runNames[4], "SSE2 Cache Test");
    sprintf(runNames[5], "3DNow Cache Test");
    sprintf(runNames[6], "SSE Memory Test");
    sprintf(runNames[7], "SSE2 Memory Test");
    sprintf(runNames[8], "3DNow Memory Test");

    keyPress = GetAsyncKeyState(VK_ESCAPE);

    // Display the window

    MoveWindow(mainHWND, xpos, ypos, 640, 480, FALSE);
    
    ShowWindow(hwnd, cmdshow);    
    UpdateWindow(hwnd);

    if (!autoRun) displayWhatever(mainHWND);
    
    return TRUE;
    
}  // InitApp


// WindowProc - handle messages for the main application window


long FAR PASCAL WindowProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam)                  
{
           
    switch( msg )
    {
    case WM_ACTIVATEAPP:
        break;
        
    case WM_INITMENUPOPUP:
        break;
        
    case WM_CREATE:
        break;

    case WM_DESTROY:
        CleanUpAndPostQuit();
        break;

    case WM_COMMAND:
        switch( LOWORD( wparam ) )
        {
        case MENU_ABOUT:
            if (testHwnd)   DestroyWindow(testHwnd);
            DialogBox( MyInstance,"AboutBox", hwnd, (DLGPROC)AboutDlgProc );
            displayWhatever(mainHWND);                                      
            break;

        case MENU_RHELP:
            if (testHwnd)   DestroyWindow(testHwnd);
            DialogBox( MyInstance,"RelHelp", hwnd, (DLGPROC)AboutDlgProc );
            displayWhatever(mainHWND);                                      
            break;

        case MENU_BHELP:
            if (testHwnd)   DestroyWindow(testHwnd);
            DialogBox( MyInstance,"BatchHelp", hwnd, (DLGPROC)AboutDlgProc );
            displayWhatever(mainHWND);                                      
            break;

        case MENU_DISPLAY:
            displayWhatever(mainHWND);                                      
            break;
        
        case MENU_EXIT:
            CleanUpAndPostQuit();
            break;
                
        }
        break;
   
        case WM_PAINT:
             break;
           
        case WM_KEYDOWN:
            break;

       case WM_ACTIVATE:
            break;
    } 
    return DefWindowProc( hwnd, msg, wparam, lparam );

} // WindowProc




// AboutDlgProc - processes messages for the about dialog.

BOOL FAR PASCAL AboutDlgProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam )
{
    lparam = lparam;

    switch( msg ) {
    case WM_INITDIALOG:
        return( TRUE );

    case WM_COMMAND:
        if( LOWORD( wparam ) == IDOK ) {
            EndDialog( hwnd, TRUE );
            return( TRUE );
        }
        break;
    }
    return( FALSE );

} // AboutDlgProc

  
void local_time()
{
    time_t time_of_day;
    time_of_day = time( NULL ); 
    strcpy (timeday, ctime(&time_of_day)); // timeday = date and time
    return;
}


// CleanUpAndPostQuit
// Release all D3DRM objects, post a quit message and set the bQuit flag

void
CleanUpAndPostQuit(void)
{
    if (started)
    {
        _aligned_free(spRand);
        _aligned_free(dpRand);
        _aligned_free(dpResults);
        _aligned_free(spResults);
    }
    bQuit = TRUE;
    return;
}


BOOL FAR PASCAL TestPanel( HWND hdlg, UINT msg,
                                WPARAM wparam, LPARAM lparam )
{
    wparam = wparam;
    lparam = lparam;
    hdlg = hdlg;
    int  k, index;
//    int FAR  tabs[11] = {1, 27, 54, 81, 108, 135, 162, 189, 216, 243, 270};
    int FAR  tabs[11] = {1, 33, 60, 87, 114, 141, 168, 195, 222, 249, 276};
    HINSTANCE hand = NULL;

     switch( msg )
     {
       case WM_INITDIALOG:
        if (hasSSE)
        {
            CreateWindow("BUTTON", "SSE", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         25, 240, 70, 25, hdlg, (HMENU)SSE_CPU, hand, NULL);  
            CreateWindow("BUTTON", "SSE", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         290, 240, 70, 25, hdlg, (HMENU)SSE_CACHE, hand, NULL);  
            CreateWindow("BUTTON", "SSE", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         550, 240, 70, 25, hdlg, (HMENU)SSE_RAM, hand, NULL);  
        }
        if (hasSSE2)
        {
            CreateWindow("BUTTON", "SSE2", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         25, 280, 70, 25, hdlg, (HMENU)SSE2_CPU, hand, NULL);  
            CreateWindow("BUTTON", "SSE2", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         290, 280, 70, 25, hdlg, (HMENU)SSE2_CACHE, hand, NULL);  
            CreateWindow("BUTTON", "SSE2", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         550, 280, 70, 25, hdlg, (HMENU)SSE2_RAM, hand, NULL);  
        }
        if (has3DNow)
        {
            CreateWindow("BUTTON", "3DNow", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         25, 320, 70, 25, hdlg, (HMENU)DNOW_CPU, hand, NULL);  
            CreateWindow("BUTTON", "3DNow", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         290, 320, 70, 25, hdlg, (HMENU)DNOW_CACHE, hand, NULL);
            CreateWindow("BUTTON", "3DNow", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                         550, 320, 70, 25, hdlg, (HMENU)DNOW_RAM, hand, NULL);  
        }
        SendDlgItemMessage(hdlg, RESULTS, LB_SETTABSTOPS, (WPARAM) 11, (LPARAM) (int FAR*) tabs);

        for (k=0; k<KBsizes; k++)
        {
            sprintf(writeMsg, "%8d",  KBCache[k]);
            SendDlgItemMessage(hdlg, AT_KB, CB_ADDSTRING, 0, (LPARAM) writeMsg);
        }
        SendDlgItemMessage(hdlg, AT_KB, CB_SETCURSEL, cacheRun, 0L);


        for (k=0; k<MBsizes; k++)
        {
            sprintf(writeMsg, "%8d", KBMemory[k]);         
            SendDlgItemMessage(hdlg, AT_MB, CB_ADDSTRING, 0, (LPARAM) writeMsg);
        }
        SendDlgItemMessage(hdlg, AT_MB, CB_SETCURSEL, RAMRun, 0L);

        for (k=0; k<MaxMins; k++)
        {
            sprintf(writeMsg, "%8d",  runTimes[k]);
            SendDlgItemMessage(hdlg, AT_MINS, CB_ADDSTRING, 0, (LPARAM) writeMsg);
        }
        SendDlgItemMessage(hdlg, AT_MINS, CB_SETCURSEL, timeToDo, 0L);
        
        SendDlgItemMessage(hdlg, RESULTS, LB_SETHORIZONTALEXTENT, 700, 0L);


        if (runDone)
        {
           SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) runMsg);
           SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) " ");
           if (errorCount == 0)
           {
               if (testFunction > 5)
               {
                   sprintf(writeMsg, " Testing time%8.2f minutes at %d MB/Sec --- No Errors     ",
                          runningSecs / 60, (int)(countMB / runningSecs));
               }
               else
               {
                   sprintf(writeMsg, " Testing time%8.2f minutes at %d MFLOPS --- No Errors     ",
                          runningSecs / 60, (int)(countMFops / runningSecs));
               }
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) " ");
               sprintf(writeMsg, "   Results are in  %s log file", logfile);
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
           }
           else
           {
               if (testFunction > 5)
               {
                   sprintf(writeMsg, " %8.2f minutes at %d MB/Sec",
                          runningSecs / 60, (int)(countMB / runningSecs));
               }
               else
               {
                   sprintf(writeMsg, " %8.2f minutes at %d MFLOPS",
                          runningSecs / 60, (int)(countMFops / runningSecs));
               }
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) " ");
               sprintf(writeMsg, "   ERRORS in test. See %s log file", logfile);
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
           }
           SendDlgItemMessage(hdlg, RESULTS, LB_SETCURSEL, 20, 0L);
        }
        else
        {
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) configData1);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) configData2);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) configData3);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) " ");
            strcpy(writeMsg, " Batch Command ");
            strncat(writeMsg, commandLine, 200);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " After running, results should be in file %s", logfile);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
        }
        return( TRUE );
                    
        case WM_COMMAND:

        if( HIWORD( wparam ) == CBN_SELCHANGE)
        {
            switch (LOWORD (wparam))
            {
                case AT_KB:
                 index = (int)SendDlgItemMessage(hdlg, AT_KB, CB_GETCURSEL, 0, 0);
                 useCacheKB = KBCache[index];
                 cacheRun = index;
                 break;                

                case AT_MB:
                 index = (int)SendDlgItemMessage(hdlg, AT_MB, CB_GETCURSEL, 0, 0);
                 useRAMKB = KBMemory[index];
                 RAMRun = index;
                 if (useRAMKB > ramKB - 32768)
                 {
                     sprintf(writeMsg, " %d KB memory selected with memory size %d KB", useRAMKB, ramKB);
                     MessageBox(NULL, writeMsg, "WARNING OF PAGING TO DISK", MB_ICONWARNING | MB_OK);
                 }
                 break;                

                case AT_MINS:
                 index = (int)SendDlgItemMessage(hdlg, AT_MINS, CB_GETCURSEL, 0, 0);
                 timeToDo = index;
                 runMins = runTimes[timeToDo];
                 break;                
            }
            
        }
        else if( HIWORD( wparam ) == BN_CLICKED)
        {
            switch (LOWORD (wparam))
            {
                case SSE_CPU:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 0;
                  isSSE2 = FALSE;
                  memKWords = 1;                                      
                  phase = 1;
                 break;

                case SSE2_CPU:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 1;
                  isSSE2 = TRUE;                                      
                  memKWords = 1;                                      
                  phase = 1;
                 break;

                case DNOW_CPU:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 2;
                  isSSE2 = FALSE;                                      
                  memKWords = 1;                                      
                  phase = 1;
                 break;
                 
                case SSE_CACHE:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 3;
                  isSSE2 = FALSE;
                  memKWords = (double)useCacheKB / 4;                                      
                  phase = 1;
                 break;

                case SSE2_CACHE:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 4;
                  isSSE2 = TRUE;
                  memKWords = (double)useCacheKB / 8;                                     
                  phase = 1;
                 break;

                case DNOW_CACHE:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 5;
                  isSSE2 = FALSE;
                  memKWords = (double)useCacheKB / 4;                                     
                  phase = 1;
                 break;

                case SSE_RAM:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 6;
                  isSSE2 = FALSE;
                  memKWords = (double)useRAMKB / 4;                                     
                  phase = 1;
                 break;

                case SSE2_RAM:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 7;
                  isSSE2 = TRUE;
                  memKWords = (double)useRAMKB / 8;                                      
                  phase = 1;
                 break;

                case DNOW_RAM:
                  doTesting = TRUE;
                  displayWhatever(mainHWND);                                      
                  testFunction = 8;
                  isSSE2 = FALSE;
                  memKWords = (double)useRAMKB / 4;                                      
                  phase = 1;
                 break;

                case EXIT_BUTTON:
                 PostMessage(mainHWND, WM_COMMAND, MENU_EXIT, 0);
                 break;
            }
        }
        return( TRUE );         
     }
     return( FALSE );
    
} // TestPanel


void start_time()
{
    QueryPerformanceCounter(&starttime);
}
 
void end_time()
{
    LARGE_INTEGER liDiff;
    LARGE_INTEGER liFreq;

    QueryPerformanceCounter(&liDiff);

    liDiff.QuadPart -= starttime.QuadPart;

    (void)QueryPerformanceFrequency(&liFreq);
    
    runningSecs = (double)liDiff.QuadPart / (double) liFreq.QuadPart;
}


BOOL setUp()
{
    HDC  hdc;
    int i, j, jr;

    hdc = GetDC(mainHWND);
    sprintf(writeMsg, "                          Allocating Memory         ");
    TextOut(hdc, 100, 100, writeMsg, lstrlen(writeMsg));
    ReleaseDC(mainHWND, hdc);
    GdiFlush();    

    outfile = fopen(logfile,"a+");
    if (outfile == NULL)
    {
        sprintf(writeMsg, "Cannot open results file");
        MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        CleanUpAndPostQuit();
        return FALSE;
    }

    if (isSSE2)
    { 
        dpData = (double *)_aligned_malloc((int)(memKWords * 8192), 16);
        if (dpData == NULL)
        {
            sprintf(writeMsg, "Cannot allocate memory");
            if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
            fprintf(outfile,"Cannot allocate memory\n\n");
            fflush(outfile);
            fclose(outfile);
            CleanUpAndPostQuit();
            return FALSE;
        }        
        for (i=0; i<(int)memKWords * 1024; i++)
        {
            dpData[i] = 0;
        } 
    }
    else
    {
        spData = (float *)_aligned_malloc((int)(memKWords * 4096), 16);
        if (spData == NULL)
        {
            sprintf(writeMsg, "Cannot allocate memory");
            if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
            fprintf(outfile,"Cannot allocate memory\n\n");
            fflush(outfile);
            fclose(outfile);
            CleanUpAndPostQuit();
            return FALSE;
        }
        for (i=0; i<(int)memKWords * 1024; i++)
        {
            spData[i] = 0;
        } 
    }
    if (testFunction < 3)
    {
        if (isSSE2)
        {
            for (i=0; i<(int)memKWords*1024; i++)
            {
               dpData[i] = dpRand[i];
            }
        }
        else
        {
            for (i=0; i<(int)memKWords*1024; i++)
            {
               spData[i] = spRand[i];
            }
        }
        
    }
    if (testFunction == 3)
    {
        jr = 0;
        for (i=0; i<(int)(memKWords*1024); i=i+128)
        {
            for (j=0; j<32; j++)
            {
                spData[i+j]    = (float)0.6 / spRand[jr+j];
            }
            for (j=0; j<32; j++)
            {
                spData[i+j+32] = spRand[jr+j] / (float)2.0;
            }
            for (j=0; j<32; j++)
            {
                spData[i+j+64] = (float)-0.7 * spRand[jr+j];
            }
            for (j=0; j<32; j++)
            {
                spData[i+j+96] = (float)1.4 * spRand[jr+j];
            }
            jr = jr + 32;
            if (jr > 1023) jr = 0;
        }
    }
    else if (testFunction == 4)
    {
        jr = 0;
        for (i=0; i<(int)(memKWords*1024); i=i+64)
        {
            for (j=0; j<16; j++)
            {
                dpData[i+j]    = 0.6 / dpRand[jr+j];
            }
            for (j=0; j<16; j++)
            {
                dpData[i+j+16] = dpRand[jr+j] / 2.0;
            }
            for (j=0; j<16; j++)
            {
                dpData[i+j+32] = -0.7 * dpRand[jr+j];
            }
            for (j=0; j<16; j++)
            {
                dpData[i+j+48] = 1.4 * dpRand[jr+j];
            }
            jr = jr + 16;
            if (jr > 1023) jr = 0;
        }
    }
    else if (testFunction == 5)
    {
        jr = 0;
        for (i=0; i<(int)(memKWords*1024); i=i+64)
        {
            for (j=0; j<16; j++)
            {
                spData[i+j]    = (float)0.6 / spRand[jr+j];
            }
            for (j=0; j<16; j++)
            {
                spData[i+j+16] = spRand[jr+j] / (float)2.0;
            }
            for (j=0; j<16; j++)
            {
                spData[i+j+32] = (float)-0.7 * spRand[jr+j];
            }
            for (j=0; j<16; j++)
            {
                spData[i+j+48] = (float)1.4 * spRand[jr+j];
            }
            jr = jr + 16;
            if (jr > 1023) jr = 0;
        }
    }
    if (testFunction < 3)
    {
        sprintf(writeMsg, " %s at %d minutes", runNames[testFunction], runMins);
        sprintf(testMsg, " %s         ", runNames[testFunction]);
    }
    else if (testFunction < 6)
    {
        sprintf(writeMsg, " %s at %d minutes and %d KB", runNames[testFunction], runMins, useCacheKB);
        sprintf(testMsg, " %s          ", runNames[testFunction]);
    }
    else
    {
        sprintf(writeMsg, " %s at %d minutes and %d KB", runNames[testFunction], runMins, useRAMKB);
        sprintf(testMsg, " %s          ", runNames[testFunction]);
    }
    SetWindowText(mainHWND, writeMsg);        
    strcpy(runMsg, writeMsg);
    strcpy(errorMsg, " ");
    local_time();

    fprintf(outfile, "\n #########################################################################\n");
    fprintf(outfile, "            SSE, SSE2 and 3DNow Reliability Test %s\n\n", version);
    fprintf(outfile, "                   Copyright (C) Roy Longbottom 2002\n\n");
    strcpy(writeMsg2, " Batch Command ");
    strncat(writeMsg2, commandLine, 200);
    fprintf(outfile, " %s\n\n", writeMsg2);
    fprintf(outfile, " %s, Start at %s\n", writeMsg, timeday);

    fprintf(outfile,"%s\n", configData1);
    fprintf(outfile,"%s\n", configData2);
    fprintf(outfile,"%s\n", startRam);        
    fprintf(outfile,"\n");
    fflush(outfile);

    firstPass = TRUE;
    runMult = 1;
    if (testFunction < 6)
    {        
        for (i=0; i<2; i++)
        {
            start_time();
            if (testFunction == 0)
            {
                SSEAM();
            }
            else if (testFunction == 1)
            {
                SSE2AM();
            }
            else if (testFunction == 2)
            {
                CPU3DNowAM();
            }
            else if (testFunction == 3)
            {
                CacheSSE1();
            }
            else if (testFunction == 4)
            {
                CacheSSE2();
            }
            else if (testFunction == 5)
            {
                Cache3DNow();
            }
            end_time();
            if (i == 0)
            {
                if (runningSecs < 0.001)
                {
                    runMult = 1024;
                }
                else
                {
                    runMult = (int)(1 / runningSecs);
                    if (runMult < 1) runMult = 1;
                }
            }
    
        }
        for (i=0; i<1024; i++)
        {
            if (isSSE2)
            {
                dpCompare[i] = dpResults[i];
            }
            else
            {
                spCompare[i] = spResults[i];
            }
        }
        if (saveData) saveNumbers();
    }
       
    keyPress = GetAsyncKeyState(VK_ESCAPE);
   
    start_time();
    loopSecs = 1.0;
    errorCount = 0;
    countMFops = 0;
    countMB = 0;
    ramTest = 0;
    ramMult = 0;
    theseSecs = 60;
    passesDone = 0;
    flash = TRUE;
    runDone = FALSE;
    hdc = GetDC(mainHWND);
    sprintf(writeMsg, "                          Running First Pass         ");
    TextOut(hdc, 100, 100, writeMsg, lstrlen(writeMsg));
    ReleaseDC(mainHWND, hdc);

    return TRUE;
} // setUp

BOOL finish()
{
    HDC  hdc;

    fprintf(outfile, "\n");
    local_time();
    fprintf(outfile, "             Reliability Test Ended %s", timeday);
    fprintf(outfile, "\n");
    fflush(outfile);
    fclose(outfile);
    doTesting = FALSE;
    if (!autoRun) displayWhatever(mainHWND);

    hdc = GetDC(mainHWND);
    Rectangle(hdc, winRect.left, winRect.top,
                   winRect.right, winRect.bottom);
    ReleaseDC(mainHWND, hdc);

    return TRUE;
    
} // finish

BOOL runTests()
{
    HDC  hdc;

    firstPass = FALSE;
    do
    {
        passesDone = passesDone + 1;
        if (testFunction == 0)
        {
            SSEAM();
        }
        else if (testFunction == 1)
        {
            SSE2AM();
        }
        else if (testFunction == 2)
        {
            CPU3DNowAM();
        }
        else if (testFunction == 3)
        {
            CacheSSE1();
        }
        else if (testFunction == 4)
        {
            CacheSSE2();
        }
        else if (testFunction == 5)
        {
            Cache3DNow();
        }
        else if (testFunction == 6)
        {
            RAMSSE1();
        }
        else if (testFunction == 7)
        {
            RAMSSE2();
        }
        else if (testFunction == 8)
        {
            RAM3DNow();
        }
        if (testFunction > 5)
        {
            if (ramTest == 0)
            {
                end_time();
                fprintf(outfile, " Pass 1 write & read %8.2f MB at %d MB/Sec\n",
                                            theseMB / 2, (int)(countMB / runningSecs));
                countMB = 0;
                start_time();
            }
            else if (ramTest == 1)
            {
                
                end_time();
                fprintf(outfile, " Pass 2 read only    %8.2f MB at %d MB/Sec\n",
                                           theseMB, (int)(countMB / runningSecs));
                if (runningSecs < 0.001)
                {
                    runMult = 1024;
                }
                else
                {
                    runMult = (int)(1 / runningSecs);
                    if (runMult < 1) runMult = 1;
                }
                countMB = 0;
                start_time();
            }
            if (ramTest == 7) ramTest = 1;
            ramTest = ramTest + 1;
        }
        end_time();
        if (runningSecs < loopSecs)
        {
            end = FALSE;
        }
        else
        {
            end = TRUE;
        }
        if (theseErrors > 3) end = TRUE;
    }
    while (!end);
    loopSecs = runningSecs + 1;

    if (theseErrors > 0)
    {
        theseErrors = 0;
        errorCount = errorCount + 1;
    }
    if (errorCount > 4)
    {
        if (testFunction > 5)
        {
            fprintf(outfile, " Testing time%8.2f minutes out of %d.0  at %d MB/Sec\n",
                          runningSecs / 60, runMins, (int)(countMB / runningSecs));
        }
        else
        {
            fprintf(outfile, " Testing time%8.2f minutes out of %d.0  at %d MFLOPS\n",
                          runningSecs / 60, runMins, (int)(countMFops / runningSecs));
        }
        fprintf(outfile, " Too many ERRORS pass %d, program exit\n", passesDone);
        fflush(outfile);
        phase = 3;
    }   
    else if (GetAsyncKeyState(VK_ESCAPE) != keyPress)
    {
        if (testFunction > 5)
        {
            fprintf(outfile, " Testing time%8.2f minutes out of %d.0  at %d MB/Sec\n",
                          runningSecs / 60, runMins, (int)(countMB / runningSecs));
        }
        else
        {
            fprintf(outfile, " Testing time%8.2f minutes out of %d.0  at %d MFLOPS\n",
                          runningSecs / 60, runMins, (int)(countMFops / runningSecs));
        }
        fprintf(outfile, " Escape Pressed, program exit\n");
        fflush(outfile);
        phase = 3;
    }
    else if (runningSecs < (double)runMins * 60)
    {
        if (errorCount == 0)
        {
            if (testFunction > 5)
            {
                sprintf(timeMsg,  "Time  %8.2f minutes       ", runningSecs / 60);
                sprintf(speedMsg, "Speed %8d MB/Sec          ", (int)(countMB / runningSecs));

                sprintf(writeMsg, " Testing time%8.2f minutes out of %d.0  at %d MB/Sec --- No Errors     ",
                                  runningSecs / 60, runMins, (int)(countMB / runningSecs));
            }
            else
            {
                sprintf(timeMsg,  "Time  %8.2f minutes       ", runningSecs / 60);
                sprintf(speedMsg, "Speed %8d MFLOPS          ", (int)(countMFops / runningSecs));
                sprintf(writeMsg, " Testing time%8.2f minutes out of %d.0  at %d MFLOPS --- No Errors     ",
                          runningSecs / 60, runMins, (int)(countMFops / runningSecs));
            }
        }
        else
        {
            if (flash)
            {
                if (testFunction > 5)
                {
                sprintf(timeMsg,  "Time  %8.2f minutes       ", runningSecs / 60);
                sprintf(speedMsg, "Speed %8d MB/Sec          ", (int)(countMB / runningSecs));
                    sprintf(writeMsg, " %8.2f minutes out of %d.0  at %d MB/Sec --- HAD ERRORS see log file   ",
                          runningSecs / 60, runMins, (int)(countMB / runningSecs));
                }
                else
                {
                sprintf(timeMsg,  "Time  %8.2f minutes       ", runningSecs / 60);
                sprintf(speedMsg, "Speed %8d MFLOPS          ", (int)(countMFops / runningSecs));
                    sprintf(writeMsg, " %8.2f minutes out of %d.0  at %d MFLOPS --- HAD ERRORS see log file   ",
                          runningSecs / 60, runMins, (int)(countMFops / runningSecs));
                }
               flash = FALSE;
            }
            else
            {
               sprintf(writeMsg, " #################################### ");
               flash = TRUE;
            } 
            sprintf(errorMsg, "See %s log file", logfile);
        }
        hdc = GetDC(mainHWND);
        if (multiWin)
        {
            TextOut(hdc, 10, 10, timeMsg, lstrlen(timeMsg));
            TextOut(hdc, 10, 23, speedMsg, lstrlen(speedMsg));
        }
        TextOut(hdc, 220, 60, testMsg, lstrlen(testMsg));
        TextOut(hdc, 80, 100, writeMsg, lstrlen(writeMsg));
        TextOut(hdc, 220, 140, errorMsg, lstrlen(errorMsg));
        sprintf(writeMsg, "Press ESC to stop program");
        TextOut(hdc, 180, 160, writeMsg, lstrlen(writeMsg));
        ReleaseDC(mainHWND, hdc);
        GdiFlush();
        if (runningSecs > theseSecs)
        {
            theseSecs = theseSecs + 60;
            if (errorCount == 0)
            {
                if (testFunction > 5)
                {
                    fprintf(outfile,"%8.2f Minutes at %d MB/Sec, No Errors\n",
                                runningSecs / 60, (int)(countMB / runningSecs));
                }
                else
                {
                    fprintf(outfile,"%8.2f Minutes at %d MFLOPS, No Errors\n",
                                runningSecs / 60, (int)(countMFops / runningSecs));
                }
                fflush(outfile);
            }
            else
            {
                if (testFunction > 5)
                {
                    fprintf(outfile,"%8.2f Minutes at %d MB/Sec, HAD ERRORS\n",
                                runningSecs / 60, (int)(countMB / runningSecs));
                }
                else
                {
                    fprintf(outfile,"%8.2f Minutes at %d MFLOPS, HAD ERRORS\n",
                                runningSecs / 60, (int)(countMFops / runningSecs));
                }                               
                fflush(outfile);
            }
        } 
    }
    else
    {
        if (errorCount == 0)
        {
            if (testFunction > 5)
            {
                fprintf(outfile,"%8.2f Minutes at %d MB/Sec, No Errors\n",
                            runningSecs / 60, (int)(countMB / runningSecs));
            }
            else
            {
                fprintf(outfile,"%8.2f Minutes at %d MFLOPS, No Errors\n",
                            runningSecs / 60, (int)(countMFops / runningSecs));
            }               
            fflush(outfile);
        }
        else
        {
            if (testFunction > 5)
            {
                fprintf(outfile,"%8.2f Minutes at %d MB/Sec, HAD ERRORS\n",
                            runningSecs / 60, (int)(countMB / runningSecs));
            }
            else
            {
                fprintf(outfile,"%8.2f Minutes at %d MFLOPS, HAD ERRORS\n",
                            runningSecs / 60, (int)(countMFops / runningSecs));
            }                              
            fflush(outfile);
        }
        phase = 3;
    }
    if (phase == 3)
    {
//        fprintf(outfile, " Finish\n");
//        fflush(outfile);   
        if (isSSE2)
        {
            _aligned_free(dpData);
//            fprintf(outfile, " Finish _aligned_free(dpData)\n");
//            fflush(outfile);
        }
        else
        {
            _aligned_free(spData);
//            fprintf(outfile, " Finish _aligned_free(spData\n");
//            fflush(outfile);    
        }
//        fprintf(outfile, " Finish _aligned_free(dpResults)\n");
//        fflush(outfile);    
    }
    return TRUE;    

} // runTests

BOOL CPU3DNowAM()
{
    int i, j, r;
    int rm = runMult * 25;
    theseMFops = (double)rm / 1000000 * 512 * 32 * 2; 
    theseErrors = 0;
    end = FALSE;

    for (j=0; j<1024; j=j+2)
    {
        for (i=0; i<2; i++)
        {
            spData[i]    = spRand[i+j];
            spData[i+2]  = (float)0.6 / spRand[i+j];
            spData[i+4]  = spRand[i+j] / (float)2.0;
            spData[i+6]  = (float)-0.7 * spRand[i+j];
            spData[i+8]  = (float)1.4 * spRand[i+j];
        }
        __asm   
        {
           femms
           align 16
           mov      edx,  rm
           mov      eax,  spData;
           movq     mm0, [eax]
           movq     mm1, mm0
           movq     mm2, mm0
           movq     mm3, mm0
           movq     mm4, [eax+8]
           movq     mm5, [eax+16]
           movq     mm6, [eax+24]
           movq     mm7, [eax+32]
           align 16
      lp1: pfmul    mm0, mm4
           pfmul    mm1, mm4
           pfmul    mm2, mm4
           pfmul    mm3, mm4
           pfmul    mm0, mm5
           pfmul    mm1, mm5
           pfmul    mm2, mm5
           pfmul    mm3, mm5   
           pfadd    mm0, mm6
           pfadd    mm1, mm6
           pfadd    mm2, mm6
           pfadd    mm3, mm6   
           pfadd    mm0, mm7
           pfadd    mm1, mm7
           pfadd    mm2, mm7
           pfadd    mm3, mm7
           pfmul    mm0, mm4
           pfmul    mm1, mm4
           pfmul    mm2, mm4
           pfmul    mm3, mm4
           pfmul    mm0, mm5
           pfmul    mm1, mm5
           pfmul    mm2, mm5
           pfmul    mm3, mm5   
           pfadd    mm0, mm6
           pfadd    mm1, mm6
           pfadd    mm2, mm6
           pfadd    mm3, mm6   
           pfadd    mm0, mm7
           pfadd    mm1, mm7
           pfadd    mm2, mm7
           pfadd    mm3, mm7
           dec      edx
           jg       lp1
           movq     [eax],    mm0
           movq     [eax+8], mm1
           movq     [eax+16], mm2
           movq     [eax+24], mm3
           femms
        }
        for (i=0; i<2; i++)
        { 
            if  (spData[i] != spData[i+2]
              || spData[i] != spData[i+4]
              || spData[i] != spData[i+6])
            {
                fprintf(outfile, " Registers different values\n");
                r = 0;
                for (i=0; i<8; i=i+2)
                {
                    fprintf(outfile, " R%d %17.7e %17.7e\n",
                               r, spData[i], spData[i+1]);
                    r = r + 1;
                }
                i = 4;
                theseErrors = theseErrors + 1;
            }
        }
        if (firstPass)
        {
            for (i=0; i<2; i++)
            {
                spResults[i+j] = spData[i];
            }
        }
        else
        {
            for (i=0; i<2; i++)
            {
                if (spCompare[i+j] != spData[i])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%4d ERROR was %17.7e should be %17.7e", i+j, spData[i], spCompare[j+i]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                }
            }
        } 
        if (theseErrors > 3)
        {
            j = 1024;
            fprintf(outfile, " End pass %d as 4 errors\n", passesDone);
            fflush(outfile);
        }        
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // CPU3DNowAM


BOOL Cache3DNow()
{
    int i, j, lp;
    int ix;
    int jr = 0;
    int jrx;
    double mk;
    
    mk = memKWords;
    
    theseErrors = 0;
    end = FALSE;
    lp = 1024 * runMult;

    theseMFops = (double)lp /1000000 * 1024 * mk;
    for (i=0; i<(int)(mk*1024); i=i+64)
    {
        ix = i * 4;
        jrx = jr * 4;
        __asm
        {
            femms
            align 16
            mov      eax,  spRand
            mov      ebx,  jrx
            add      eax,  ebx
            movq     mm0, [eax]
            movq     mm1, [eax+8]
            movq     mm2, [eax+16]
            movq     mm3, [eax+24]
            movq     mm4, [eax+32]
            movq     mm5, [eax+40]
            movq     mm6, [eax+48]
            movq     mm7, [eax+56]
            mov      eax,  spData
            mov      ebx,  ix
            add      eax,  ebx
            mov      edx,  lp
            align 16
      lp1:  pfmul    mm0, [eax]
            pfmul    mm1, [eax+8]
            pfmul    mm2, [eax+16]
            pfmul    mm3, [eax+24]
            pfmul    mm4, [eax+32]
            pfmul    mm5, [eax+40]
            pfmul    mm6, [eax+48]
            pfmul    mm7, [eax+56]
            pfmul    mm0, [eax+64]
            pfmul    mm1, [eax+72]
            pfmul    mm2, [eax+80]
            pfmul    mm3, [eax+88]
            pfmul    mm4, [eax+96]
            pfmul    mm5, [eax+104]
            pfmul    mm6, [eax+112]
            pfmul    mm7, [eax+120]
            pfadd    mm0, [eax+128]
            pfadd    mm1, [eax+136]
            pfadd    mm2, [eax+144]
            pfadd    mm3, [eax+152]
            pfadd    mm4, [eax+160]
            pfadd    mm5, [eax+168]
            pfadd    mm6, [eax+176]
            pfadd    mm7, [eax+184]
            pfadd    mm0, [eax+192]
            pfadd    mm1, [eax+200]
            pfadd    mm2, [eax+208]
            pfadd    mm3, [eax+216]
            pfadd    mm4, [eax+224]
            pfadd    mm5, [eax+232]
            pfadd    mm6, [eax+240]
            pfadd    mm7, [eax+248]             
            dec      edx
            jg       lp1
            mov      eax,  spResults
            mov      ebx,  jrx
            add      eax,  ebx
            movq     [eax],    mm0
            movq     [eax+8],  mm1
            movq     [eax+16], mm2
            movq     [eax+24], mm3
            movq     [eax+32], mm4
            movq     [eax+40], mm5
            movq     [eax+48], mm6
            movq     [eax+56], mm7
            femms
        }
        if (!firstPass)
        {
            for (j=0; j<16; j++)
            {
                if (spResults[jr+j] != spCompare[jr+j])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%d ERROR was %17.7e should be %17.7e",
                                        i+j, spResults[jr+j], spCompare[jr+j]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                    if (theseErrors > 31)
                    {
                        i = (int)mk*1024;
                        j = 32;
                        fprintf(outfile, " End pass %d as 32 errors\n", passesDone);
                        fflush(outfile);
                    }
                }
            }
        }
        jr = jr + 16;
        if (jr > 1023) jr = 0;
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // Cache3DNow

BOOL RAM3DNow()
{
    int i;
    int count, count2;
    int j;
    int jj;
    double mk;
    
    mk = memKWords;
    theseMB = runMult * mk * 4 * 1024 /1000000;
    count = (int)(mk * 4 * 1024 / 64);
    count2 = count / 2;
    
    theseErrors = 0;
    end = FALSE;
    j = ramMult*4;
    if (ramTest == 0 || ramTest == 7)
    {
        ramMult = ramMult + 16;
        if (ramMult > 1023) ramMult = 0;
        theseMB = theseMB + (mk * 4 * 1024 /1000000);
        __asm
        {
            femms      
            align 16
            mov      eax,  spRand
            mov      edx,  j
            add      eax,  edx
            movq     mm0, [eax]
            movq     mm1, [eax+8]
            movq     mm2, [eax+16]
            movq     mm3, [eax+24]
            movq     mm4, [eax+32]
            movq     mm5, [eax+40]
            movq     mm6, [eax+48]
            movq     mm7, [eax+56]
            mov      eax, spData
            mov      edx, count
            align 16
       lp1: movq     [eax],     mm0
            movq     [eax+8],   mm1
            movq     [eax+16],  mm2
            movq     [eax+24],  mm3
            movq     [eax+32],  mm4
            movq     [eax+40],  mm5
            movq     [eax+48],  mm6
            movq     [eax+56],  mm7
            add      eax, 64
            dec      edx
            jg       lp1
            mov      eax,  spData
            movq     mm0, [eax]
            movq     mm1, [eax+8]
            movq     mm2, [eax+16]
            movq     mm3, [eax+24]
            movq     mm4, [eax+32]
            movq     mm5, [eax+40]
            movq     mm6, [eax+48]
            movq     mm7, [eax+56]
            mov      eax, spRand
            mov      edx,  j
            add      eax,  edx
            pfsub    mm0, [eax]
            pfsub    mm1, [eax+8]
            pfsub    mm2, [eax+16]
            pfsub    mm3, [eax+24]
            pfsub    mm4, [eax+32]
            pfsub    mm5, [eax+40]
            pfsub    mm6, [eax+48]
            pfsub    mm7, [eax+56]
            pfadd    mm0, mm1
            pfadd    mm0, mm2
            pfadd    mm0, mm3
            pfadd    mm0, mm4
            pfadd    mm0, mm5
            pfadd    mm0, mm6
            pfadd    mm0, mm7
            mov      eax,  spResults
            movq   [eax], mm0
            femms      
        }
        for (i=0; i<2; i++)
        {
            if (spResults[i] != 0.0)
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Initial ERROR was %17.7e should be %17.7e",
                                     spResults[i], 0.0);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
        }
    }
    spResults[0] = 0;
    spResults[1] = 0;

    __asm
    {
        femms      
        align 16
        mov      eax,  spResults
        movq     mm0, [eax]
        movq     mm1, mm0
        movq     mm2, mm0
        movq     mm3, mm0
        movq     mm4, mm0
        movq     mm5, mm0
        movq     mm6, mm0
        movq     mm7, mm0
        mov      ecx, runMult
   lp3: mov      eax,  spData
        mov      edx,  count2
        align 16
   lp2: pfadd    mm0, [eax]
        pfadd    mm1, [eax+8]
        pfadd    mm2, [eax+16]
        pfadd    mm3, [eax+24]
        pfadd    mm4, [eax+32]
        pfadd    mm5, [eax+40]
        pfadd    mm6, [eax+48]
        pfadd    mm7, [eax+56]
        pfsub    mm0, [eax+64]
        pfsub    mm1, [eax+72]
        pfsub    mm2, [eax+80]
        pfsub    mm3, [eax+88]
        pfsub    mm4, [eax+96]
        pfsub    mm5, [eax+104]
        pfsub    mm6, [eax+112]
        pfsub    mm7, [eax+120]
        add      eax, 128
        dec      edx
        jg       lp2
        dec      ecx
        jg       lp3
        pfadd    mm0, mm1
        pfadd    mm0, mm2
        pfadd    mm0, mm3
        pfadd    mm0, mm4
        pfadd    mm0, mm5
        pfadd    mm0, mm6
        pfadd    mm0, mm7
        mov      eax,  spResults
        movq   [eax], mm0
        femms      
    } 
    for (i=0; i<2; i++)
    {
        if (spResults[i] != 0.0)
        {
            theseErrors = theseErrors + 1;
            sprintf (writeMsg, " Sumcheck ERROR was %17.7e should be %17.7e",
                                 spResults[i], 0.0);
            fprintf(outfile, "%s\n", writeMsg);
            fflush(outfile);
        }
    }
    if (theseErrors  > 0)
    {
        jj = 0;
        for (i=0; i<(int)mk*1024; i++)
        {
            if (spData[i] != spRand[j+jj])
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Word %10d Compare ERROR was %17.7e should be %17.7e",
                                     i, spData[i], spRand[j+jj]);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
            jj = jj + 1;
            if (jj == 16) jj = 0;
            if (theseErrors > 7)
            {
                fprintf(outfile, " End pass %d as 8 errors\n", ramTest+1);
                fflush(outfile);
                i = (int)mk*1024;
            }
        }
    }
    countMB = countMB + theseMB;
    return TRUE;    

} // RAM3DNow


BOOL SSEAM()
{
    int i, j, r;
    int rm = runMult * 25;
    theseMFops = (double)rm / 1000000 * 256 * 32 * 4; 
    theseErrors = 0;
    end = FALSE;

    for (j=0; j<1024; j=j+4)
    {
        for (i=0; i<4; i++)
        {
            spData[i]    = spRand[i+j];
            spData[i+4]  = (float)0.6 / spRand[i+j];
            spData[i+8]  = spRand[i+j] / (float)2.0;
            spData[i+12] = (float)-0.7 * spRand[i+j];
            spData[i+16] = (float)1.4 * spRand[i+j];
        }
        __asm   
        {
           align 16
           mov      edx,  rm
           mov      eax,  spData;
           movaps   xmm0, [eax]
           movaps   xmm1, xmm0
           movaps   xmm2, xmm0
           movaps   xmm3, xmm0
           movaps   xmm4, [eax+16]
           movaps   xmm5, [eax+32]
           movaps   xmm6, [eax+48]
           movaps   xmm7, [eax+64]
           align 16
      lp1: mulps    xmm0, xmm4
           mulps    xmm1, xmm4
           mulps    xmm2, xmm4
           mulps    xmm3, xmm4
           mulps    xmm0, xmm5
           mulps    xmm1, xmm5
           mulps    xmm2, xmm5
           mulps    xmm3, xmm5   
           addps    xmm0, xmm6
           addps    xmm1, xmm6
           addps    xmm2, xmm6
           addps    xmm3, xmm6   
           addps    xmm0, xmm7
           addps    xmm1, xmm7
           addps    xmm2, xmm7
           addps    xmm3, xmm7
           mulps    xmm0, xmm4
           mulps    xmm1, xmm4
           mulps    xmm2, xmm4
           mulps    xmm3, xmm4
           mulps    xmm0, xmm5
           mulps    xmm1, xmm5
           mulps    xmm2, xmm5
           mulps    xmm3, xmm5   
           addps    xmm0, xmm6
           addps    xmm1, xmm6
           addps    xmm2, xmm6
           addps    xmm3, xmm6   
           addps    xmm0, xmm7
           addps    xmm1, xmm7
           addps    xmm2, xmm7
           addps    xmm3, xmm7
           dec      edx
           jg       lp1
           movaps   [eax],    xmm0
           movaps   [eax+16], xmm1
           movaps   [eax+32], xmm2
           movaps   [eax+48], xmm3
        }
        for (i=0; i<4; i++)
        { 
            if  (spData[i] != spData[i+4]
              || spData[i] != spData[i+8]
              || spData[i] != spData[i+12])
            {
                fprintf(outfile, " Registers different values\n");
                r = 0;
                for (i=0; i<16; i=i+4)
                {
                    fprintf(outfile, " R%d %17.7e  %17.7e  %17.7e  %17.7e\n",
                               r, spData[i], spData[i+1], spData[i+2], spData[i+3]);
                    r = r + 1;
                }
                i = 4;
                theseErrors = theseErrors + 1;
            }
        }
        if (firstPass)
        {
            for (i=0; i<4; i++)
            {
                spResults[i+j] = spData[i];
            }
        }
        else
        {
            for (i=0; i<4; i++)
            {
                if (spCompare[i+j] != spData[i])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%4d ERROR was %17.7e should be %17.7e", i+j, spData[i], spCompare[j+i]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                }
            }
        } 
        if (theseErrors > 7)
        {
            j = 1024;
            fprintf(outfile, " End pass %d as 8 errors\n", passesDone);
            fflush(outfile);
        }        
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // SSEAM


BOOL CacheSSE1()
{
    int i, j, lp;
    int ix;
    int jr = 0;
    int jrx;
    double mk;

    mk = memKWords;
    
    theseErrors = 0;
    end = FALSE;
    lp = 1024 * runMult;

    theseMFops = (double)lp /1000000 * 1024 * mk;
    for (i=0; i<(int)(mk*1024); i=i+128)
    {
        ix = i * 4;
        jrx = jr * 4;
        __asm
        {
            align 16
            mov      eax,  spRand
            mov      ebx,  jrx
            add      eax,  ebx
            movaps   xmm0, [eax]
            movaps   xmm1, [eax+16]
            movaps   xmm2, [eax+32]
            movaps   xmm3, [eax+48]
            movaps   xmm4, [eax+64]
            movaps   xmm5, [eax+80]
            movaps   xmm6, [eax+96]
            movaps   xmm7, [eax+112]
            mov      eax,  spData
            mov      ebx,  ix
            add      eax,  ebx
            mov      edx,  lp
            align 16
      lp1:  mulps    xmm0, [eax+0]
            mulps    xmm1, [eax+16]
            mulps    xmm2, [eax+32]
            mulps    xmm3, [eax+48]
            mulps    xmm4, [eax+64]
            mulps    xmm5, [eax+80]
            mulps    xmm6, [eax+96]
            mulps    xmm7, [eax+112]
            mulps    xmm0, [eax+128]
            mulps    xmm1, [eax+144]
            mulps    xmm2, [eax+160]
            mulps    xmm3, [eax+176]
            mulps    xmm4, [eax+192]
            mulps    xmm5, [eax+208]
            mulps    xmm6, [eax+224]
            mulps    xmm7, [eax+240]
            addps    xmm0, [eax+256]
            addps    xmm1, [eax+272]
            addps    xmm2, [eax+288]
            addps    xmm3, [eax+304]
            addps    xmm4, [eax+320]
            addps    xmm5, [eax+336]
            addps    xmm6, [eax+352]
            addps    xmm7, [eax+368]
            addps    xmm0, [eax+384]
            addps    xmm1, [eax+400]
            addps    xmm2, [eax+416]
            addps    xmm3, [eax+432]
            addps    xmm4, [eax+448]
            addps    xmm5, [eax+464]
            addps    xmm6, [eax+480]
            addps    xmm7, [eax+496]      
            dec      edx
            jg       lp1
            mov      eax,  spResults
            mov      ebx,  jrx
            add      eax,  ebx
            movaps   [eax],    xmm0
            movaps   [eax+16], xmm1
            movaps   [eax+32], xmm2
            movaps   [eax+48], xmm3
            movaps   [eax+64], xmm4
            movaps   [eax+80], xmm5
            movaps   [eax+96], xmm6
            movaps   [eax+112], xmm7
        }
        if (!firstPass)
        {
            for (j=0; j<32; j++)
            {
                if (spResults[jr+j] != spCompare[jr+j])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%d ERROR was %17.7e should be %17.7e",
                                        i+j, spResults[jr+j], spCompare[jr+j]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                    if (theseErrors > 31)
                    {
                        i = (int)mk*1024;
                        j = 32;
                        fprintf(outfile, " End pass %d as 32 errors\n", passesDone);
                        fflush(outfile);
                    }
                }
            }
        }
        jr = jr + 32;
        if (jr > 1023) jr = 0;
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // CacheSSE1

BOOL RAMSSE1()
{
    int i;
    int count, count2;
    int j = 0;
    int jj;
    double mk;
    
    mk = memKWords;
    theseMB = runMult * mk * 4 * 1024 /1000000;
    count = (int)(mk * 4 * 1024 / 128);
    count2 = count / 2;
    
    theseErrors = 0;
    end = FALSE;
    j = ramMult*4;
    if (ramTest == 0 || ramTest == 7)
    {
        ramMult = ramMult + 32;
        if (ramMult > 1023) ramMult = 0;
        theseMB = theseMB + (mk * 4 * 1024 /1000000);
        __asm
        {
            align 16
            mov      eax,  spRand
            mov      edx,  j
            add      eax,  edx
            movaps   xmm0, [eax]
            movaps   xmm1, [eax+16]
            movaps   xmm2, [eax+32]
            movaps   xmm3, [eax+48]
            movaps   xmm4, [eax+64]
            movaps   xmm5, [eax+80]
            movaps   xmm6, [eax+96]
            movaps   xmm7, [eax+112]
            mov      eax,  spData
            mov      edx,  count
            align 16
       lp1: movaps   [eax],     xmm0
            movaps   [eax+16],  xmm1
            movaps   [eax+32],  xmm2
            movaps   [eax+48],  xmm3
            movaps   [eax+64],  xmm4
            movaps   [eax+80],  xmm5
            movaps   [eax+96],  xmm6
            movaps   [eax+112], xmm7
            add      eax, 128
            dec      edx
            jg       lp1
            mov      eax,  spData
            movaps   xmm0, [eax]
            movaps   xmm1, [eax+16]
            movaps   xmm2, [eax+32]
            movaps   xmm3, [eax+48]
            movaps   xmm4, [eax+64]
            movaps   xmm5, [eax+80]
            movaps   xmm6, [eax+96]
            movaps   xmm7, [eax+112]
            mov      eax,  spRand
            mov      edx,  j
            add      eax,  edx
            subps    xmm0, [eax]
            subps    xmm1, [eax+16]
            subps    xmm2, [eax+32]
            subps    xmm3, [eax+48]
            subps    xmm4, [eax+64]
            subps    xmm5, [eax+80]
            subps    xmm6, [eax+96]
            subps    xmm7, [eax+112]
            addps    xmm0, xmm1
            addps    xmm0, xmm2
            addps    xmm0, xmm3
            addps    xmm0, xmm4
            addps    xmm0, xmm5
            addps    xmm0, xmm6
            addps    xmm0, xmm7
            mov      eax,  spResults
            movaps   [eax], xmm0
        }
        for (i=0; i<4; i++)
        {
            if (spResults[i] != 0.0)
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Initial ERROR was %17.7e should be %17.7e",
                                     spResults[i], 0.0);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
        }
    }
    spResults[0] = 0;
    spResults[1] = 0;
    spResults[2] = 0;
    spResults[3] = 0;

    __asm
    {
        align 16
        mov      eax,  spResults
        movaps   xmm0, [eax]
        movaps   xmm1, xmm0
        movaps   xmm2, xmm0
        movaps   xmm3, xmm0
        movaps   xmm4, xmm0
        movaps   xmm5, xmm0
        movaps   xmm6, xmm0
        movaps   xmm7, xmm0
        mov      ecx, runMult
   lp3: mov      eax,  spData
        mov      edx,  count2
        align 16
   lp2: addps    xmm0, [eax]
        addps    xmm1, [eax+16]
        addps    xmm2, [eax+32]
        addps    xmm3, [eax+48]
        addps    xmm4, [eax+64]
        addps    xmm5, [eax+80]
        addps    xmm6, [eax+96]
        addps    xmm7, [eax+112]
        subps    xmm0, [eax+128]
        subps    xmm1, [eax+144]
        subps    xmm2, [eax+160]
        subps    xmm3, [eax+176]
        subps    xmm4, [eax+192]
        subps    xmm5, [eax+208]
        subps    xmm6, [eax+224]
        subps    xmm7, [eax+240]
        add      eax, 256
        dec      edx
        jg       lp2
        dec      ecx
        jg       lp3
        addps    xmm0, xmm1
        addps    xmm0, xmm2
        addps    xmm0, xmm3
        addps    xmm0, xmm4
        addps    xmm0, xmm5
        addps    xmm0, xmm6
        addps    xmm0, xmm7
        mov      eax,  spResults
        movaps   [eax], xmm0
    } 
    for (i=0; i<4; i++)
    {
        if (spResults[i] != 0.0)
        {
            theseErrors = theseErrors + 1;
            sprintf (writeMsg, " Sumcheck ERROR was %17.7e should be %17.7e",
                                 spResults[i], 0.0);
            fprintf(outfile, "%s\n", writeMsg);
            fflush(outfile);
        }
    }
    if (theseErrors  > 0)
    {
        jj = 0;
        for (i=0; i<(int)mk*1024; i++)
        {
            if (spData[i] != spRand[j+jj])
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Word %10d Compare ERROR was %17.7e should be %17.7e",
                                     i, spData[i], spRand[j+jj]);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
            jj = jj + 1;
            if (jj == 32) jj = 0;
            if (theseErrors > 7)
            {
                fprintf(outfile, " End pass %d as 8 errors\n", ramTest+1);
                fflush(outfile);
                i = (int)mk*1024;
            }
        }
    }
    countMB = countMB + theseMB;
    return TRUE;    

} // RAMSSE1

BOOL SSE2AM()
{
    int i, j, r;
    int rm = runMult * 25;
    theseMFops = (double)rm / 1000000 * 512 * 32 * 2; 
    theseErrors = 0;
    end = FALSE;

    for (j=0; j<1024; j=j+2)
    {
        for (i=0; i<2; i++)
        {
            dpData[i]    = dpRand[i+j];
            dpData[i+2]  = (float)0.6 / dpRand[i+j];
            dpData[i+4]  = dpRand[i+j] / (float)2.0;
            dpData[i+6]  = (float)-0.7 * dpRand[i+j];
            dpData[i+8]  = (float)1.4 * dpRand[i+j];
        }
        __asm   
        {
           align 16
           mov      edx,  rm
           mov      eax,  dpData;
           movapd   xmm0, [eax]
           movapd   xmm1, xmm0
           movapd   xmm2, xmm0
           movapd   xmm3, xmm0
           movapd   xmm4, [eax+16]
           movapd   xmm5, [eax+32]
           movapd   xmm6, [eax+48]
           movapd   xmm7, [eax+64]
           align 16
      lp1: mulpd    xmm0, xmm4
           mulpd    xmm1, xmm4
           mulpd    xmm2, xmm4
           mulpd    xmm3, xmm4
           mulpd    xmm0, xmm5
           mulpd    xmm1, xmm5
           mulpd    xmm2, xmm5
           mulpd    xmm3, xmm5   
           addpd    xmm0, xmm6
           addpd    xmm1, xmm6
           addpd    xmm2, xmm6
           addpd    xmm3, xmm6   
           addpd    xmm0, xmm7
           addpd    xmm1, xmm7
           addpd    xmm2, xmm7
           addpd    xmm3, xmm7
           mulpd    xmm0, xmm4
           mulpd    xmm1, xmm4
           mulpd    xmm2, xmm4
           mulpd    xmm3, xmm4
           mulpd    xmm0, xmm5
           mulpd    xmm1, xmm5
           mulpd    xmm2, xmm5
           mulpd    xmm3, xmm5   
           addpd    xmm0, xmm6
           addpd    xmm1, xmm6
           addpd    xmm2, xmm6
           addpd    xmm3, xmm6   
           addpd    xmm0, xmm7
           addpd    xmm1, xmm7
           addpd    xmm2, xmm7
           addpd    xmm3, xmm7
           dec      edx
           jg       lp1
           movapd   [eax],    xmm0
           movapd   [eax+16], xmm1
           movapd   [eax+32], xmm2
           movapd   [eax+48], xmm3
        }
        for (i=0; i<2; i++)
        { 
            if  (dpData[i] != dpData[i+2]
              || dpData[i] != dpData[i+4]
              || dpData[i] != dpData[i+6])
            {
                fprintf(outfile, " Registers different values\n");
                r = 0;
                for (i=0; i<8; i=i+2)
                {
                    fprintf(outfile, " R%d %25.15e %25.15e\n",
                               r, dpData[i], dpData[i+1]);
                    r = r + 1;
                }
                i = 4;
                theseErrors = theseErrors + 1;
            }
        }
        if (firstPass)
        {
            for (i=0; i<2; i++)
            {
                dpResults[i+j] = dpData[i];
            }
        }
        else
        {
            for (i=0; i<2; i++)
            {
                if (dpCompare[i+j] != dpData[i])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%4d ERROR was %25.15e should be %25.15e", i+j, dpData[i], dpCompare[j+i]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                }
            }
        } 
        if (theseErrors > 3)
        {
            j = 1024;
            fprintf(outfile, " End pass %d as 4 errors\n", passesDone);
            fflush(outfile);
        }        
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // SSE2AM


BOOL CacheSSE2()
{
    int i, j, lp;
    int ix;
    int jr = 0;
    int jrx;
    double mk;
    
    mk = memKWords;

    theseErrors = 0;
    end = FALSE;
    lp = 1024 * runMult;

    theseMFops = (double)lp /1000000 * 1024 * mk;
    for (i=0; i<(int)(mk*1024); i=i+64)
    {
        ix = i * 8;
        jrx = jr * 8;
        __asm
        {
            align 16
            mov      eax,  dpRand
            mov      ebx,  jrx
            add      eax,  ebx
            movapd   xmm0, [eax]
            movapd   xmm1, [eax+16]
            movapd   xmm2, [eax+32]
            movapd   xmm3, [eax+48]
            movapd   xmm4, [eax+64]
            movapd   xmm5, [eax+80]
            movapd   xmm6, [eax+96]
            movapd   xmm7, [eax+112]
            mov      eax,  dpData
            mov      ebx,  ix
            add      eax,  ebx
            mov      edx,  lp
            align 16
      lp1:  mulpd    xmm0, [eax+0]
            mulpd    xmm1, [eax+16]
            mulpd    xmm2, [eax+32]
            mulpd    xmm3, [eax+48]
            mulpd    xmm4, [eax+64]
            mulpd    xmm5, [eax+80]
            mulpd    xmm6, [eax+96]
            mulpd    xmm7, [eax+112]
            mulpd    xmm0, [eax+128]
            mulpd    xmm1, [eax+144]
            mulpd    xmm2, [eax+160]
            mulpd    xmm3, [eax+176]
            mulpd    xmm4, [eax+192]
            mulpd    xmm5, [eax+208]
            mulpd    xmm6, [eax+224]
            mulpd    xmm7, [eax+240]
            addpd    xmm0, [eax+256]
            addpd    xmm1, [eax+272]
            addpd    xmm2, [eax+288]
            addpd    xmm3, [eax+304]
            addpd    xmm4, [eax+320]
            addpd    xmm5, [eax+336]
            addpd    xmm6, [eax+352]
            addpd    xmm7, [eax+368]
            addpd    xmm0, [eax+384]
            addpd    xmm1, [eax+400]
            addpd    xmm2, [eax+416]
            addpd    xmm3, [eax+432]
            addpd    xmm4, [eax+448]
            addpd    xmm5, [eax+464]
            addpd    xmm6, [eax+480]
            addpd    xmm7, [eax+496]      
            dec      edx
            jg       lp1
            mov      eax,  dpResults
            mov      ebx,  jrx
            add      eax,  ebx
            movapd   [eax],    xmm0
            movapd   [eax+16], xmm1
            movapd   [eax+32], xmm2
            movapd   [eax+48], xmm3
            movapd   [eax+64], xmm4
            movapd   [eax+80], xmm5
            movapd   [eax+96], xmm6
            movapd   [eax+112], xmm7
        }
        if (!firstPass)
        {
            for (j=0; j<16; j++)
            {
                if (dpResults[jr+j] != dpCompare[jr+j])
                {
                    theseErrors = theseErrors + 1;
                    sprintf (writeMsg, " W%d ERROR was %25.15e should be %25.15e",
                                        i+j, dpResults[jr+j], dpCompare[jr+j]);
                    fprintf(outfile, "%s\n", writeMsg);
                    fflush(outfile);
                    if (theseErrors > 15)
                    {
                        i = (int)mk*1024;
                        j = 16;
                        fprintf(outfile, " End pass %d as 32 errors\n", passesDone);
                        fflush(outfile);
                    }
                }
            }
        }
        jr = jr + 16;
        if (jr > 1023) jr = 0;
    }
    countMFops = countMFops + theseMFops;
    return TRUE;    

} // CacheSSE2

BOOL RAMSSE2()
{
    int i;
    int count, count2;
    int j = 0;
    int jj;
    double mk;
    
    mk = memKWords;
    theseMB = runMult * mk * 8 * 1024 /1000000;
    count = (int)(mk * 8 * 1024 / 128);
    count2 = count / 2;
    
    theseErrors = 0;
    end = FALSE;
    j = ramMult*4;
    if (ramTest == 0 || ramTest == 7)
    {
        ramMult = ramMult + 16;
        if (ramMult > 1023) ramMult = 0;
        theseMB = theseMB + (mk * 8 * 1024 /1000000);
        __asm
        {
            align 16
            mov      eax,  dpRand
            mov      edx,  j
            add      eax,  edx
            movapd   xmm0, [eax]
            movapd   xmm1, [eax+16]
            movapd   xmm2, [eax+32]
            movapd   xmm3, [eax+48]
            movapd   xmm4, [eax+64]
            movapd   xmm5, [eax+80]
            movapd   xmm6, [eax+96]
            movapd   xmm7, [eax+112]
            mov      eax,  dpData
            mov      edx,  count
            align 16
       lp1: movapd   [eax],     xmm0
            movapd   [eax+16],  xmm1
            movapd   [eax+32],  xmm2
            movapd   [eax+48],  xmm3
            movapd   [eax+64],  xmm4
            movapd   [eax+80],  xmm5
            movapd   [eax+96],  xmm6
            movapd   [eax+112], xmm7
            add      eax, 128
            dec      edx
            jg       lp1
            mov      eax,  dpData
            movapd   xmm0, [eax]
            movapd   xmm1, [eax+16]
            movapd   xmm2, [eax+32]
            movapd   xmm3, [eax+48]
            movapd   xmm4, [eax+64]
            movapd   xmm5, [eax+80]
            movapd   xmm6, [eax+96]
            movapd   xmm7, [eax+112]
            mov      eax,  dpRand
            mov      edx,  j
            add      eax,  edx
            subpd    xmm0, [eax]
            subpd    xmm1, [eax+16]
            subpd    xmm2, [eax+32]
            subpd    xmm3, [eax+48]
            subpd    xmm4, [eax+64]
            subpd    xmm5, [eax+80]
            subpd    xmm6, [eax+96]
            subpd    xmm7, [eax+112]
            addpd    xmm0, xmm1
            addpd    xmm0, xmm2
            addpd    xmm0, xmm3
            addpd    xmm0, xmm4
            addpd    xmm0, xmm5
            addpd    xmm0, xmm6
            addpd    xmm0, xmm7
            mov      eax,  dpResults
            movapd   [eax], xmm0
        }
        for (i=0; i<2; i++)
        {
            if (dpResults[i] != 0.0)
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Initial ERROR was %17.7e should be %17.7e",
                                     dpResults[i], 0.0);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
        }
    }
    dpResults[0] = 0;
    dpResults[1] = 0;


    __asm
    {
        align 16
        mov      eax,  dpResults
        movapd   xmm0, [eax]
        movapd   xmm1, xmm0
        movapd   xmm2, xmm0
        movapd   xmm3, xmm0
        movapd   xmm4, xmm0
        movapd   xmm5, xmm0
        movapd   xmm6, xmm0
        movapd   xmm7, xmm0
        mov      ecx,  runMult
   lp3: mov      eax,  dpData
        mov      edx,  count2
        align 16
   lp2: addpd    xmm0, [eax]
        addpd    xmm1, [eax+16]
        addpd    xmm2, [eax+32]
        addpd    xmm3, [eax+48]
        addpd    xmm4, [eax+64]
        addpd    xmm5, [eax+80]
        addpd    xmm6, [eax+96]
        addpd    xmm7, [eax+112]
        subpd    xmm0, [eax+128]
        subpd    xmm1, [eax+144]
        subpd    xmm2, [eax+160]
        subpd    xmm3, [eax+176]
        subpd    xmm4, [eax+192]
        subpd    xmm5, [eax+208]
        subpd    xmm6, [eax+224]
        subpd    xmm7, [eax+240]
        add      eax, 256
        dec      edx
        jg       lp2
        dec      ecx
        jg       lp3
        addpd    xmm0, xmm1
        addpd    xmm0, xmm2
        addpd    xmm0, xmm3
        addpd    xmm0, xmm4
        addpd    xmm0, xmm5
        addpd    xmm0, xmm6
        addpd    xmm0, xmm7
        mov      eax,  dpResults
        movapd   [eax], xmm0
    } 
    for (i=0; i<2; i++)
    {
        if (dpResults[i] != 0.0)
        {
            theseErrors = theseErrors + 1;
            sprintf (writeMsg, " Sumcheck ERROR was %17.7e should be %17.7e",
                                 dpResults[i], 0.0);
            fprintf(outfile, "%s\n", writeMsg);
            fflush(outfile);
        }
    }
    if (theseErrors  > 0)
    {
        jj = 0;
        for (i=0; i<(int)mk*1024; i++)
        {
            if (dpData[i] != dpRand[j+jj])
            {
                theseErrors = theseErrors + 1;
                sprintf (writeMsg, " Word %10d Compare ERROR was %17.7e should be %17.7e",
                                     i, dpData[i], dpRand[j+jj]);
                fprintf(outfile, "%s\n", writeMsg);
                fflush(outfile);
            }
            jj = jj + 1;
            if (jj == 32) jj = 0;
            if (theseErrors > 7)
            {
                fprintf(outfile, " End pass %d as 8 errors\n", ramTest+1);
                fflush(outfile);
                i = (int)mk*1024;
            }
        }
    }
    countMB = countMB + theseMB;
    return TRUE;    

} // RAMSSE2


BOOL saveNumbers()
{
    FILE *numfile;
    int i; 
    
    numfile = fopen("numbers.txt","a+");
    if (numfile == NULL)
    {
        sprintf(writeMsg, "Cannot open numbers file");
        MessageBox(NULL, writeMsg, "ERROR", MB_ICONWARNING | MB_OK);
        return FALSE;
    }
    if (saveRand)
    {
        for (i=0; i<1024; i=i+4)
        {
            fprintf(numfile, " %17.7e %17.7e %17.7e %17.7e\n",
                       spRand[i], spRand[i+1], spRand[i+2], spRand[i+3]);
        }
        fprintf(numfile, "\n");
        for (i=0; i<1024; i=i+4)
        {
            fprintf(numfile, " %25.15e %25.15e %25.15e %25.15e\n",
                             dpRand[i], dpRand[i+1], dpRand[i+2], dpRand[i+3]);
        }
        fprintf(numfile, "\n");
    }
    if (saveData)
    {
        if (isSSE2)
        {
            for (i=0; i<1024; i=i+4)
            {
                fprintf(numfile, " %25.15e %25.15e %25.15e %25.15e\n",
                                 dpCompare[i], dpCompare[i+1], dpCompare[i+2], dpCompare[i+3]);
            }
            fprintf(numfile, "\n");
        }
        else
        {
            for (i=0; i<1024; i=i+4)
            {
                fprintf(numfile, " %17.7e %17.7e %17.7e %17.7e\n",
                                 spCompare[i], spCompare[i+1], spCompare[i+2], spCompare[i+3]);
            }
            fprintf(numfile, "\n");
        }
    }
    fclose(numfile);
    return TRUE;
}

// sprintf (writeMsg, " %d %15.2f %15.2f", cp, memKWords, theseMFops);    
// MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);


